/** New code for algebraic fitting of an ellipse
	without solving the generalized eigenvalues problem.
	This version calculates the eigenvalues and
	eigenvectors of a symmetric 3x3 matrix

	Fraga 23/04/2010
**/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "dsyevv3.h"

void imprime_matriz( double a[][3], char *s )
{	   
	int i, j;


	printf ("--- %s----\n", s );
	for (i=0; i<3; i++){
		for (j=0; j<3; j++)
			printf ("%g ", a[i][j]);
		printf ("\n"); 
	}
}

/* 
   Realiza C = A * B
*/
void multiplicaAB(double A[][3], double B[][3], double C[][3])
{
	int i, j;
	for (i=0; i<3;i++)
		for (j=0;j<3;j++)
			C[i][j] = A[i][0]*B[0][j] + A[i][1]*B[1][j] + A[i][2]*B[2][j];
}

/* 
   Realiza C = A * B^T
*/
void multiplicaABT(double A[][3], double B[][3], double C[][3])
{   
	int i, j; 
	for (i=0; i<3;i++)
		for (j=0;j<3;j++)
			C[i][j] = A[i][0]*B[j][0] + A[i][1]*B[j][1] + A[i][2]*B[j][2];
}

int main( int argc, char *argv[] )
{

	int i,j,k;
	int num_puntos;
	char namein[64];
	double px, py, pxy, px2, py2;

	double eje_a, eje_b;
	double angulo;
	double aux1,aux2,aux3;

	double a,b,cc,d,f,g;

	FILE *fname;

	double S1[3][3], S2[3][3], S3[3][3];
	double Z[3][3], T[3][3], M[3][3];
	double Vec[3][3], Val[3];

	for( j=0; j<3; j++ )
		for( k=0; k<3; k++ ) S1[j][k] = S2[j][k] = S3[j][k] = 0;

	if ( argc != 2 ) {
		fprintf( stderr, "Args: file_name\n" ); 
		exit(1);
	}

	strncpy( namein, argv[1], 64 );
	num_puntos = 0;
	fname  = fopen (namein,"r");

	for (i=0; ;) {
		k = fscanf( fname,"%lf %lf\n", &px, &py );

		if ( k == EOF ) break;
		else if( k==2 ) {
			printf( "%lf %lf\n", px, py );

			pxy = px*py;
			px2 = px*px;
			py2 = py*py;

			Z[0][0] = px2*px2;
			Z[0][1] = Z[1][0] = px2*pxy;
			Z[0][2] = Z[1][1] = Z[2][0] = pxy*pxy;
			Z[1][2] = Z[2][1] = pxy*py2;
			Z[2][2] = py2*py2;

			// Suma
			for( j=0; j<3; j++ )
				for( k=0; k<3; k++ ) S1[j][k] += Z[j][k];

			Z[0][0] = px*px2;
			Z[0][1] = Z[1][0] = px*pxy;
			Z[0][2] = px2;
			Z[1][1] = Z[2][0] = pxy*py;
			Z[1][2] = pxy;
			Z[2][1] = py*py2;
			Z[2][2] = py2;

			// Suma
			for( j=0; j<3; j++ )
				for( k=0; k<3; k++ ) S2[j][k] += Z[j][k];

			Z[0][0] = px2;
			Z[0][1] = Z[1][0] = pxy;
			Z[0][2] = Z[2][0] = px;
			Z[1][1] = py2;
			Z[1][2] = Z[2][1] = py;
			Z[2][2] = 1.0;

			// Suma
			for( j=0; j<3; j++ )
				for( k=0; k<3; k++ ) S3[j][k] += Z[j][k];

			i++; 
		}
	}
	num_puntos = i;
	printf( "# n = %d\n", num_puntos );

	// Calcular la inversa de S3
	// Copia de S3
	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ ) M[i][j] = S3[i][j];

	// Calcular los eigenvalores y eigenvectores de S3
	if ( dsyevv3( M, Vec, Val) < 0 ){
		fprintf( stderr, "Error calculating eigenvectors\n" );
		return -1;
	}
	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ ) M[i][j] = 0.0;

	M[0][0] = 1.0/Val[0];
	M[1][1] = 1.0/Val[1];
	M[2][2] = 1.0/Val[2];


	// S3^(-1) = T * transpose(Eigvec)
	//	   T = Eigvec * M
	multiplicaAB( Vec, M, T );
	multiplicaABT( T, Vec, M );  // M = S3^(-1)
	imprime_matriz( M, "S3i" );

	multiplicaABT( M, S2, T );  // T = S3^(-1) * transpose(S2)
	// T la voy a usar depus 
	multiplicaAB( S2, T, M );   // M = S2 * T = S2 * S3^(-1) * S2'
	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ ) Z[i][j] = S1[i][j] - M[i][j];

	// Change 16/May/2011. 4 lines
	a = 10.0/Z[2][2];
	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ ) 
			Z[i][j] *= a;

	imprime_matriz( Z, "Z" );

	if ( dsyevv3( Z, M, Val) < 0 ){
		fprintf( stderr, "Error calculating eigenvectors\n" );
		return -1;
	}
	printf("Eigen1: %g %g %g\n", Val[0], Val[1], Val[2] );

	// Change 16/May/2011. 3 lines
	for( i=0; i<3; i++ ){ 
		if ( Val[i] < 0.0 ) Val[i] *= -1.0;
	}
	printf("Eigen1: %lf %lf %lf\n", Val[0], Val[1], Val[2] );

	a = 1.0/sqrt(Val[0]);
	b  = 1.0/sqrt(Val[1]);
	cc = 1.0/sqrt(Val[2]);
	printf("abc: %lf %lf %lf\n", a, b, cc );

	// Calculo de D * M' * C * M * D 
	Z[0][0] = a*a*(4.0*M[0][0]*M[2][0] - M[1][0]*M[1][0] );
	Z[0][1] = Z[1][0] =  a*b*(2.0*M[0][0]*M[2][1] + 2.0*M[0][1]*M[2][0]-M[1][0]*M[1][1]);
	Z[0][2] = Z[2][0] = a*cc*(2.0*M[0][0]*M[2][2] + 2.0*M[0][2]*M[2][0]-M[1][0]*M[1][2]);

	Z[1][1] = b*b*(4.0*M[0][1]*M[2][1] - M[1][1]*M[1][1] );
	Z[1][2] = Z[2][1] = b*cc*(2.0*M[0][1]*M[2][2] + 2.0*M[0][2]*M[2][1]-M[1][1]*M[1][2]);

	Z[2][2] = cc*cc*(4.0*M[0][2]*M[2][2] - M[1][2]*M[1][2] );

	if ( dsyevv3( Z, Vec, Val) < 0 ){
		fprintf( stderr, "Error calculating eigenvectors\n" );
		return -1;
	}

	for( i=0; i<3; i++ ){ 
		M[i][0] *= a;
		M[i][1] *= b;
		M[i][2] *= cc;
	}
	multiplicaAB( M, Vec, Z );
	

	// cond = 4*Z(1,:) .* Z(3,:) - Z(2,:) .^ 2;
	// a1 = Z(:, find(cond>0));
	// a = [ a1; -T*a1 ]

	for( i=0; i<3; i++ ) 
		Val[i] = 4.0 * Z[0][i] * Z[2][i] - Z[1][i]*Z[1][i];

	j = 4;
	for( i=0; i<3; i++ ) {
		if( Val[i] > 0 ) {
			j = i;
			break;
		}
	}
	if( j==4 ) {
		printf( "Error: these points do not define an ellipse\n" );
		return 1;
	}
	a = Z[0][j];
	b = Z[1][j]/2.0;
	cc = Z[2][j];

	for( i=0; i<3; i++ ) Val[i] = -Z[i][j]; 

	for( i=0; i<3; i++ )
		M[i][0] = Val[0]*T[i][0] + Val[1]*T[i][1] + Val[2]*T[i][2];
	
	d = M[0][0]/2.0;
	f = M[1][0]/2.0;
	g = M[2][0];

	printf( "%lf %lf %lf %lf %lf %lf\n", a, b*2.0, cc, M[0][0], M[1][0], M[2][0] );
	
	angulo =  atan (  2.0 * b / (a - cc) )*90.0/M_PI;

	aux3 = b*b - a*cc;
	aux1 = (cc*d - b*f)/aux3;
	aux2 = (a*f - b*d)/aux3;

	printf("\nCentro X =%lf\n", aux1 );
	printf("Centro Y =%lf\n", aux2 );

	aux1 = fabs( 2.0 * ( a*f*f+ cc*d*d + g*b*b - 2.0*b*d*f - a*cc*g ) );
	aux2 = fabs( (b*b-a*cc) * ( (cc-a)*sqrt(1.0 + ( (4.0*b*b)/((a-cc)*(a-cc)) ) ) - (cc+a) ) );
	aux3 = fabs( (b*b-a*cc) * ( (a-cc)*sqrt(1.0 + ( (4.0*b*b)/((a-cc)*(a-cc)) ) ) - (cc+a) ) );

	eje_a = sqrt ( aux1 / aux2);
	eje_b = sqrt ( aux1 / aux3);
	
	if( eje_a < eje_b ) {
		aux1 = eje_a;
		eje_a = eje_b;
		eje_b = aux1;
		if( angulo < 0 ) {
			angulo += 90.0;
		}
		else {
			angulo -= 90.0;
		}
	}

	printf("Semieje a : %lf \n", eje_a);
	printf("Semieje b : %lf \n", eje_b);
	printf("Angulo %lf\n",  angulo);

	return 0;
}

